/*

  procinfo.c

  Displays general info from /proc.

  Date:        1995-07-08 05:30:34
  Last Change: 2004-06-17 11:28:13

  Copyright (c) 1994-2001 svm@kozmix.cistron.nl
  Copyright (c) 2004 procinfo@meisel.cc (upgrades for 2.6 kernel)

  This software is released under the GNU Public Licence. See the file
  `COPYING' for details. Since you're probably running Linux I'm sure
  your hard disk is already infested with copies of this file, but if
  not, mail me and I'll send you one.

*/

static char *rcsid __attribute__ ((unused)) = "$Id: procinfo.c,v 1.56 2001/02/25 11:29:13 svm Exp svm $";

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <ncurses/termcap.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/param.h>	/* for HZ -- should be in <time.h> ? */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/utsname.h>

#ifdef DEBUG
#define DMALLOC_FUNC_CHECK
#include <dmalloc.h>
#endif

#include "procinfo.h"

void init_terminal_data (void);
char *my_tgets (char *te);

char *cd;			/* clear to eos */
char *ce;			/* clear to eol */
char *cl;			/* clear screen */
char *cm;			/* move cursor */
char *ho;			/* home */
char *se;			/* standout off */
char *so;			/* standout on */
char *ve;			/* cursor on */
char *vi;			/* cursor off */
int co;				/* columns */
int li;				/* lines */
int sg;				/* cookie width */

int fs = 0, redrawn = 0;
int show_stat = 0;
int show_moddev = 0;
int show_sys = 0;
int show_all = 0;
int show_diff = 0;
int show_new_mem = 0;
int show_real_mem = 0;
int show_from_baseline = 0;
int io_or_blk = 0;
int linux_version_code;
int per_sec = 0;
int squeeze_irqs = 1;
int nr_irqs = 0;
int nr_cpus = 0;

FILE *loadavgfp, *meminfofp, *modulesfp, *statfp, *uptimefp, *diskstatsfp,
    *devicesfp, *filesystemsfp, *interruptsfp, *dmafp, *cmdlinefp,
    *versionfp, *cpuinfofp, *vmstatfp;

char line[1024], cmdline[1024], booted[40], *version = NULL, *message = NULL;

float rate = 1.0;		/* per interval or per sec */

struct info new, old, base;
struct utsname ut;

unsigned long sys_hz, usr_hz;

void get_HZ(void)
{
    FILE* hzfp;

    if (!access (PROC_DIR "sys/kernel/HZ", R_OK)) {
	hzfp = myfopen (PROC_DIR "sys/kernel/HZ");
	fscanf (hzfp, "%li", &sys_hz);
	fclose (hzfp);
    } else if (linux_version_code >= 0x20560) {
	struct timespec ts;
	if (clock_getres(CLOCK_REALTIME, &ts) < 0)
	    sys_hz = 1000;
	else {
	    double res = ts.tv_sec+(((double)ts.tv_nsec)/1.0e9);
	    sys_hz = (unsigned long)((1.0/res)+0.5);
	}
    } else
	sys_hz = 100;
    if ((long)(usr_hz = (unsigned long)sysconf(_SC_CLK_TCK)) == -1)
#ifdef HZ
	usr_hz = HZ;
#else
	usr_hz = 100;
#endif
}

static void
first_page (long sl)
{
    unsigned long elapsed;
    char loadavg[128];
    int i, havetwosix = 0, lastdisk;
    static int have_m_c = -1;	/* Do we have cache info? */
    static int have_m_l = -1;	/* Do we have new-style-one-datum-per-line? */
    char devicename[128];
    char *last;
    unsigned int q,w,a,s,d,f,g,h,j,k,l,y,x;
    int disk_counter = 0, ret;


/**** /proc/uptime ****/

    fclose (uptimefp); uptimefp = myfopen (PROC_DIR "uptime");

    fgets (line, sizeof (line), uptimefp);
    new.uptime =
	(unsigned long) (atof (strtok (line, " ")) * (unsigned long) usr_hz);


/**** /proc/meminfo ****/

    fclose (meminfofp); meminfofp = myfopen (PROC_DIR "meminfo");

    if (have_m_l == -1 || have_m_c == -1) {
	char kickme[1024];
	/* God, this next stuff needs AT LEAST a macro to hide behind. */
	fclose (meminfofp); meminfofp = myfopen (PROC_DIR "meminfo");
	fread (kickme, 1024, 1, meminfofp);
	if (strstr (kickme, "MemTotal:"))
	    have_m_l = 1;
	else
	    have_m_l = 0;
	if (strstr (kickme, "Cached:"))
	    have_m_c = 1;
	else
	    have_m_c = 0;
	fclose (meminfofp); meminfofp = myfopen (PROC_DIR "meminfo");
    }

    printf ("Memory:      Total        Used        Free      "
	    "Shared     Buffers      %s\n", have_m_c ? "Cached" : "");

    fgets (line, sizeof (line), meminfofp);

    if (have_m_l) {

	new.m_to = new.m_us = new.m_fr = new.m_sh = new.m_bu = 0;
	new.s_to = new.s_us = new.s_fr = new.m_bo = new.m_an = 0;
	new.n_fs = new.s_lr = new.s_lu = new.s_ca = 0;

	do {
	    char *type = strtok (line, ":");

	    if (ISSTR ("MemTotal"))
		new.m_to = VAL;
	    else if (ISSTR ("MemFree"))
		new.m_fr = VAL;
	    else if (ISSTR ("MemShared"))
		new.m_sh = VAL;
	    else if (ISSTR ("Buffers"))
		new.m_bu = VAL;
	    else if (ISSTR ("Cached"))
		new.m_ca = VAL;
	    else if (ISSTR ("SwapTotal"))
		new.s_to = VAL;
	    else if (ISSTR ("SwapFree"))
		new.s_fr = VAL;
	    else if (ISSTR ("SwapCached"))
		new.s_ca = VAL;
	    else if (ISSTR ("Bounce"))
		new.m_bo = VAL;
	    else if (ISSTR ("AnonPages"))
		new.m_an = VAL;
	    else if (ISSTR ("NFS_Unstable"))
		new.n_fs = VAL;
	    else if (ISSTR ("SReclaimable"))
		new.s_lr = VAL;
	    else if (ISSTR ("SUnreclaim"))
		new.s_lu = VAL;
	} while (fgets (line, sizeof (line), meminfofp));

	new.m_us = new.m_to - new.m_fr;
	new.s_us = new.s_to - new.s_fr;
	new.m_ca += new.s_lr;
	new.m_ca += new.s_lu;
	new.m_ca += new.n_fs;
	new.m_ca += new.s_ca;

    } else {

	fgets (line, sizeof (line), meminfofp);
	strtok (line, " ");
	new.m_to = VAL / 1024;
	new.m_us = VAL / 1024;
	new.m_fr = VAL / 1024;
	new.m_sh = VAL / 1024;
	new.m_bu = VAL / 1024;
	if (have_m_c)
	    new.m_ca = VAL / 1024;
    }

    if (show_new_mem) {
	printf ("Mem:  %12ld%12ld%12ld%12ld%12ld",
		new.m_to, new.m_us, new.m_fr, new.m_sh,
		new.m_bu);
	if (have_m_c)
	    printf ("%12ld\n", new.m_ca);
	else
	    putchar ('\n');
    } else {
	/*
	  We're going to cheat a bit by replacing rate with sl/1e6 in
	  the DIFF macro, so that we'll end up with nice multiples of
	  4 when using the -d option.
	*/
#define rate (sl>1000000 ? (float)(sl/1000000) : (float)1.0)
	printf ("Mem:  %12ld%12ld%12ld%12ld%12ld",
		DIFF (m_to), DIFF (m_us), DIFF (m_fr), DIFF (m_sh), DIFF (m_bu));
	if (have_m_c)
	    printf ("%12ld\n", DIFF (m_ca));
	else
	    putchar ('\n');
#undef rate
    }

    if (have_m_l) {
	/* Taken care of above... */
    } else {
    	fgets (line, sizeof (line), meminfofp);
    	strtok (line, " ");
    	new.s_to = VAL / 1024;
    	new.s_us = VAL / 1024;
    	new.s_fr = VAL / 1024;
    }

    if (show_new_mem) {
        if (show_real_mem)
	    printf ("-/+ buffers%s %11ld%12ld\n",
                    have_m_c ? "/cache:" : ":      ",
		    new.m_us - new.m_bu - new.m_ca,
		    new.m_fr + new.m_bu + new.m_ca);

	printf ("Swap: %12ld%12ld%12ld\n\n",
		new.s_to, new.s_us, new.s_fr);
    } else {
	if (show_real_mem)
	    printf ("-/+ buffers%s %11ld%12ld\n",
	    	    have_m_c ? "/cache:" : ":      ",
		    DIFF (m_us) - DIFF (m_bu) - DIFF (m_ca),
		    DIFF (m_fr) + DIFF (m_bu) + DIFF (m_ca));

	printf ("Swap: %12ld%12ld%12ld\n\n",
		DIFF (s_to), DIFF (s_us), DIFF (s_fr));
    }


/**** /proc/loadavg ****/

    fseek (loadavgfp, 0L, SEEK_SET);
    fgets (loadavg, sizeof (loadavg), loadavgfp);
    if ((last = strchr(loadavg, '\n')))
	*last = '\0';

    printf ("Bootup: %s    Load average: %s\n\n", booted, loadavg);

/**** /proc/vmstat ... 2.6.x kernel only ****/
	if (vmstatfp) {
		havetwosix = 1;
		fseek (vmstatfp, 0L, SEEK_SET);
		while (fgets (line, sizeof (line), vmstatfp)) {
			char *type = strtok (line, " ");
			if (ISSTR ("pgpgin")) {
				new.pgin = VAL;
			} else if (ISSTR ("pgpgout")) {
				new.pgout = VAL;
			} else if (ISSTR ("pgactivate")) {
				new.pgac = VAL;
			} else if (ISSTR ("pgdeactivate")) {
				new.pgdeac = VAL;
			} else if (ISSTR ("pgfault")) {
				new.pgfault = VAL;
			} else if (ISSTR ("pswpin")) {
				new.swin = VAL;
			} else if (ISSTR ("pswpout")) {
				new.swout = VAL;
			}
		}
	}

/**** /proc/diskstats ... 2.6.x kernel only ****/
   if (diskstatsfp) {
	havetwosix = 1;
	fseek (diskstatsfp, 0L, SEEK_SET);
	while (fgets (line, sizeof (line), diskstatsfp)) {
	   char *type = strtok (line, "\n");
		   ret = sscanf (type, "%d%d%127s%d%d%d%d%d%d%d%d%d%d%d", &q, &w, devicename, &a, &s, &d, &f, &g, &h, &j, &k, &l, &y, &x);
		   if (ret != 14) /* Everything with less than 14 elements is not a disk device */
			   continue;
         last = devicename+strlen(devicename)-1;
         if (isalpha(*last)) {/* Is the last char is a number ignore it .. like fd0 */
            if (++disk_counter > 7) /* 7 disks is all we have room for */
               continue;
            else {
               new.disk_r[disk_counter-1] = a;
               new.disk_w[disk_counter-1] = g;
            }
         }
      }
   }

/**** /proc/stat ****/

    fseek (statfp, 0L, SEEK_SET);
    while (fgets (line, sizeof (line), statfp)) {
	char *type = strtok (line, " ");

	if (ISSTR ("cpu")) {
	    new.cpu_user = VAL;
	    new.cpu_nice = VAL;
	    new.cpu_sys = VAL;
	    new.cpu_idle = VAL;
	    new.cpu_iowait = VAL;
	    new.cpu_hirq = VAL;
	    new.cpu_sirq = VAL;
	    new.cpu_steal = VAL;
	    new.cpu_guest = VAL;
	    /*
	     * according to bug #1959, sometimes the cpu_idle
	     * seems to go backwards(!) on SMP boxes.  This may
	     * have been fixed in later kernel releases.  This
	     * is just a sanity check.
	     */
	    if (new.cpu_idle < old.cpu_idle)
	      new.cpu_idle = old.cpu_idle;
	} else if (ISSTR ("disk")) {
	    new.disk[0] = VAL;
	    new.disk[1] = VAL;
	    new.disk[2] = VAL;
	    new.disk[3] = VAL;
	    new.disk[4] = VAL;
	    new.disk[5] = VAL;
	    new.disk[6] = VAL;
	} else if (ISSTR ("disk_rio") && io_or_blk == 0) {
	    new.disk_r[0] = VAL;
	    new.disk_r[1] = VAL;
	    new.disk_r[2] = VAL;
	    new.disk_r[3] = VAL;
	    new.disk_r[4] = VAL;
	    new.disk_r[5] = VAL;
	    new.disk_r[6] = VAL;
	} else if (ISSTR ("disk_wio") && io_or_blk == 0) {
	    new.disk_w[0] = VAL;
	} else if (ISSTR ("disk_wio") && io_or_blk == 0) {
	    new.disk_w[0] = VAL;
	} else if (ISSTR ("disk_wio") && io_or_blk == 0) {
	    new.disk_w[0] = VAL;
	} else if (ISSTR ("disk_wio") && io_or_blk == 0) {
	    new.disk_w[0] = VAL;
	    new.disk_w[1] = VAL;
	    new.disk_w[2] = VAL;
	    new.disk_w[3] = VAL;
	    new.disk_w[4] = VAL;
	    new.disk_w[5] = VAL;
	    new.disk_w[6] = VAL;
	} else if (ISSTR ("disk_rblk") && io_or_blk == 1) {
	    new.disk_r[0] = VAL;
	    new.disk_r[1] = VAL;
	    new.disk_r[2] = VAL;
	    new.disk_r[3] = VAL;
	    new.disk_r[4] = VAL;
	    new.disk_r[5] = VAL;
	    new.disk_r[6] = VAL;
	} else if (ISSTR ("disk_wblk") && io_or_blk == 1) {
	    new.disk_w[0] = VAL;
	    new.disk_w[1] = VAL;
	    new.disk_w[2] = VAL;
	    new.disk_w[3] = VAL;
	    new.disk_w[4] = VAL;
	    new.disk_w[5] = VAL;
	    new.disk_w[6] = VAL;
	} else if (ISSTR ("disk_io:")) {
	    unsigned int q, w, e, r, t, y, u; /* I'm NOT in the mood today. */
	    char *barf;

	    while ((barf = strtok (NULL, " "))) {

		if (++disk_counter > 7) /* 7 is all we have room for */
		    continue;

		ret = sscanf (barf, "(%d,%d):(%d,%d,%d,%d,%d)",
			      &q, &w, &e, &r, &t, &y, &u);
		if (ret != 7)
		    continue;

		if (io_or_blk == 1) {
		    new.disk_r[disk_counter-1] = t;
		    new.disk_w[disk_counter-1] = u;
		} else {
		    new.disk_r[disk_counter-1] = r;
		    new.disk_w[disk_counter-1] = y;
		}
	    }
	} else if (ISSTR ("page")) {
	    new.pgin = VAL;
	    new.pgout = VAL;
	} else if (ISSTR ("swap")) {
	    new.swin = VAL;
	    new.swout = VAL;
	} else if (ISSTR ("intr")) {
	    if (nr_irqs) {
		size_t len = strlen(&line[5]);
		char *tok, *lp = &line[5+len-1];

		VAL;		/* First value is total of all interrupts,
				   for compatibility with rpc.rstatd. We
				   ignore it. */

		tok = (char*)0;
		for (i = 0; i < nr_irqs; i++) {
		    char * cp = strtok(tok, " ");
		    if (!cp)
			break;
		    new.intr[i] = strtoul(cp, NULL, 10);
		    tok = (char*)0;

		    if (*lp != '\n') {
			off_t off = cp - line + strlen(cp) + 1;
			if (off + sizeof(line)/4 > sizeof(line)) {
			    size_t rest = strlen(&line[off]);
			    memmove((tok = &line[0]), &line[off], rest+1);
			    if (!fgets(&line[rest], sizeof(line) - rest, statfp)) {
				*lp = '\n';
				continue;
			    }
			    len = strlen(&line[rest]);
			    lp = &line[rest+len-1];
			}
		    }
		}
	    } else
		new.old_intr = VAL;
	} else if (ISSTR ("ctxt")) {
	    new.ctxt = VAL;
	} else if (ISSTR ("syscalls")) {
	    new.syscalls = VAL;
	}
    }

    elapsed = new.uptime;

/* XXX Is this stuff still relevant/true? */

#ifdef __i386__		/* IRQ 0 is timer tick on i386's... */
    if (nr_irqs && new.intr[0] && (new.intr[0] != old.intr[0])) {
	if (fs && old.uptime)
	    elapsed = (DIFF(intr[0]) * usr_hz) / sys_hz;
    } else
#endif
#ifdef __sparc__	/* IRQ 10 is timer tick on sparc's... */
    if (nr_irqs && new.intr[10] && (new.intr[0] != old.intr[0])) {
	if (fs && old.uptime)
	    elapsed = (DIFF(intr[10]) * usr_hz) / sys_hz;
    } else
#endif
    {
	/* This won't be exact... */
	if (fs && old.uptime)
	    elapsed = DIFF (uptime);
    }

    printf ("user  : %s %s",
	    hms (bDIFF (cpu_user)), perc (bDIFF (cpu_user), elapsed, nr_cpus));
    printf ("  page in :%11lu", bDIFF (pgin));
    if (new.disk_r[0])
	printf ("  disk 1: %8lur%8luw\n", bDIFF (disk_r[0]),
		bDIFF (disk_w[0]));
    else if (new.disk[0])
	printf ("  disk 1: %8lu\n", bDIFF (disk[0]));
    else
	putchar ('\n');

    printf ("nice  : %s %s",
	    hms (bDIFF (cpu_nice)), perc (bDIFF (cpu_nice), elapsed, nr_cpus));
    printf ("  page out:%11lu", bDIFF (pgout));
    if (new.disk_r[1])
	printf ("  disk 2: %8lur%8luw\n", bDIFF (disk_r[1]),
		bDIFF (disk_w[1]));
    else if (new.disk[1])
	printf ("  disk 2: %8lu\n", bDIFF (disk[1]));
    else
	putchar ('\n');

    printf ("system: %s %s",
	    hms (bDIFF (cpu_sys)), perc (bDIFF (cpu_sys), elapsed, nr_cpus));
    if (havetwosix)
        printf ("  page act:%11lu", bDIFF (pgac));
    else
        printf ("  swap in :%11lu", bDIFF (swin));
    if (new.disk_r[2])
	printf ("  disk 3: %8lur%8luw\n", bDIFF (disk_r[2]),
		bDIFF (disk_w[2]));
    else if (new.disk[2])
	printf ("  disk 3: %8lu\n", bDIFF (disk[2]));
    else
	putchar ('\n');

    if (havetwosix) {
        printf ("IOwait: %s %s",
            hms (bDIFF (cpu_iowait)), perc (bDIFF (cpu_iowait), elapsed, nr_cpus));
        printf ("  page dea:%11lu", bDIFF (pgdeac));
        if (new.disk_r[3])
	    printf ("  disk 4: %8lur%8luw\n", bDIFF (disk_r[3]),
		bDIFF (disk_w[3]));
        else if (new.disk[3])
	    printf ("  disk 4: %8lu\n", bDIFF (disk[3]));
        else
	    putchar ('\n');

        printf ("hw irq: %s %s",
            hms (bDIFF (cpu_hirq)), perc (bDIFF (cpu_hirq), elapsed, nr_cpus));
        printf ("  page flt:%11lu", bDIFF (pgfault));
        if (new.disk_r[4])
	    printf ("  disk 5: %8lur%8luw\n", bDIFF (disk_r[4]),
		bDIFF (disk_w[4]));
        else if (new.disk[4])
	    printf ("  disk 5: %8lu\n", bDIFF (disk[4]));
        else
	    putchar ('\n');

        printf ("sw irq: %s %s",
            hms (bDIFF (cpu_sirq)), perc (bDIFF (cpu_sirq), elapsed, nr_cpus));
        printf ("  swap in :%11lu", bDIFF (swin));
        if (new.disk_r[5])
	    printf ("  disk 6: %8lur%8luw\n", bDIFF (disk_r[5]),
		bDIFF (disk_w[5]));
        else if (new.disk[5])
	    printf ("  disk 6: %8lu\n", bDIFF (disk[5]));
        else
	    putchar ('\n');
    }

    printf ("idle  : %s %s",
	    hms (bDIFF (cpu_idle)), perc (bDIFF (cpu_idle), elapsed, nr_cpus));
    printf ("  swap out:%11lu", bDIFF (swout));
    if (havetwosix)
        lastdisk = 6;
    else
        lastdisk = 3;
    if (new.disk_r[lastdisk])
	printf ("  disk %d: %8lur%8luw\n", lastdisk+1, bDIFF (disk_r[lastdisk]),
		bDIFF (disk_w[lastdisk]));
    else if (new.disk[lastdisk])
	printf ("  disk %d: %8lu\n", lastdisk+1, bDIFF (disk[lastdisk]));
    else
	putchar ('\n');

    if (havetwosix) {
	if (new.cpu_steal) {
	    printf ("steal : %s %s",
		hms (bDIFF (cpu_steal)), perc (bDIFF (cpu_steal), elapsed, nr_cpus));
	    putchar ('\n');
	}
	if (new.cpu_guest) {
	    printf ("guest : %s %s",
		hms (bDIFF (cpu_guest)), perc (bDIFF (cpu_guest), elapsed, nr_cpus));
	    putchar ('\n');
	}
    }

    printf ("uptime: %s         context :%11lu", hms (new.uptime),
	    bDIFF (ctxt));
    if (new.syscalls)	/* If we have this, we can use the old interrupts spot. */
	printf ("  syscalls: %16lu", bDIFF (syscalls));

/****= /proc/interrupts =****/

    if (nr_irqs) {
	char emtpy[1] = { '\0' };
	struct irq_s { char *label; } irq[nr_irqs];

	for (i = 0; i < nr_irqs; i++)
	    irq[i].label = &emtpy[0];

	if (interruptsfp) {
	    int i;

	    fseek (interruptsfp, 0L, SEEK_SET);
	    while (fgets (line, sizeof (line), interruptsfp)) {
		char *q, *p;
		int n;

		if (!(q = strchr(line, ':')))		/* Skip CPU0 ... */
		    continue;
		p = &line[0];
		*q = '\0';

		while (*p == ' ')
		   p++;

		if ((*p < '0') || (*p > '9'))		/* Skip NMI, LOC, ... */
		    continue;

		i = atol(p);				/* Get the IRQ no. */

		if (i >= nr_irqs)			/* Should not happen */
		    continue;

		q++;
		while (*q == ' ')			/* Eat white spaces */
		    q++;

		n = 0;
		while (n < nr_cpus) {
		    if (*q == '\0' || *q == '\n')
			break;
		    p = q;
		    if (!(q = strchr(p, ' ')))		/* Skip no. of IRQ */
			break;
		    while (*q == ' ')			/* Eat white spaces */
			q++;
		    n++;
		}					/* Do this for all CPUs */

		if (!q || (*q == '\0') || (*q == '\n'))
		    continue;

		if (!(p = strchr(q, ' ')))		/* Skip type of IRQ */
		    continue;

		while (*p == ' ')			/* Eat white spaces */
		    p++;

		if ((q = strchr(p, '\n')))		/* Eat newline character */
		    *q = '\0';

		if (*(irq[i].label) == 0) {
		    irq[i].label = strndupa(p, 21);
		    if (irq[i].label == (char*)0)
			irq[i].label = &emtpy[0];
		}
	    }
	}

/**** /proc/dma ****/

	if (dmafp) {
	    int i;
	    size_t tmplen;
	    char tmp[22];

	    fseek (dmafp, 0L, SEEK_SET);
	    while (fgets (line, sizeof (line), dmafp)) {
		char *q, *p;
		size_t len;
		long num;

		q = strchr(line, ':');
		if (!q || (*q == '\0') || (*q == '\n'))
		    continue;

		p = &line[0];
		*q = '\0';

		while (*p == ' ')			/* Eat white spaces */
		    p++;
		num = atol(p);

		q++;
		while (*q == ' ')			/* Eat white spaces */
		    q++;

		if ((p = strchr(q, '\n')))		/* Eat newline character */
		    *p = '\0';

		if ((p = strchr(q, ' ')))		/* Forget rest of line */
		    *p = '\0';
		len = strlen(q);

		for (i = 0; i < nr_irqs; i++) {
		    struct irq_s *s = &irq[i];

		    if (*(s->label) == '\0')
			continue;

		    if (strncmp (q, s->label, len) == 0) {
			tmplen = snprintf(tmp, sizeof(tmp), " [%ld]", num);

			if (tmplen > 21)		/* Should not happen */
			    continue;
			tmp[tmplen] = '\0';

			if (strlen(s->label) > (21 - tmplen))
			    s->label[21 - tmplen] = 0;

			strcat (s->label, tmp);
		    }
		}
	    }
	}

	fputs ("\n\n", stdout);

	if (squeeze_irqs) {
	    int howmany = 0, rows;
	    struct {
		int nr;
		unsigned long count;
		char *label;
	    } squirqs[nr_irqs];

	    for (i = 0; i < nr_irqs; i++) {
		struct irq_s *s = &irq[i];
		if (new.intr[i] || *(s->label)) {
		    squirqs[howmany].nr = i;
		    squirqs[howmany].count = bDIFF(intr[i]);
		    squirqs[howmany].label = s->label;
		    howmany++;
		}
	    }

	    rows = (howmany + 1) / 2;

	    for (i = 0; i < rows; i++) {
		/* The last row may be incomplete if howmany is odd, hence: */
		if (i == rows - 1 && howmany & 1) {
		    printf ("irq%4d:%10lu %-20s\n",
			    squirqs[i].nr, squirqs[i].count, squirqs[i].label);
		} else {
		    printf ("irq%4d:%10lu %-20s "
			    "irq%4d:%10lu %-20s\n",
			    squirqs[i].nr, squirqs[i].count, squirqs[i].label,
			    squirqs[i+rows].nr,
			    squirqs[i+rows].count,
			    squirqs[i+rows].label);
		}
	    }
	} else {
	    for (i = 0; i < nr_irqs / 2; i++) {
		int j = i+(nr_irqs/2);
		struct irq_s *s = &irq[i];
		struct irq_s *r = &irq[j];
		printf ("irq%3d:%10lu %-21s "
			"irq%3d:%10lu %-21s\n",
			i, bDIFF (intr[i]), s->label,
			j, bDIFF (intr[j]), r->label);
	    }
	}

    } else
	printf ("\tinterrupts: %8lu\n", bDIFF (old_intr));

}

static void
second_page (void)
{
    if (show_all)
	putchar ('\n');


/**** /proc/cmdline ****/

    /*
      We're not fclosing cmdlinefp because we're using it later as a
      check.
    */

    fgets (cmdline, 1024, cmdlinefp);

    if (cmdlinefp)
	printf ("Kernel Command Line:\n  %s\n", cmdline);


/**** /proc/modules ****/

    if (modulesfp) {
	int barf = 0;

	fseek (modulesfp, 0L, SEEK_SET);
	printf ("Modules:\n");
	while (fgets (line, sizeof (line), modulesfp)) {
	    char *mod;
	    long pg;
	    int used = 0;

	    if (co - barf < 20 ) {
		barf = 0;
		putchar('\n');
	    }
	    barf += 20;

	    mod = strtok (line, " ");
	    if (strlen (mod) > 14)
		mod[14] = '\0';
	    if (linux_version_code < 0x20100)
		pg = VAL * 4;			/* size is in 4kb pages */
	    else
		pg = (VAL + 512) / 1024;	/* size is in bytes */
	    used = VAL;
	    printf ("%3ld %c%-14s ",
		    pg, used ? '*' : ' ', mod);
	}

	/*if (co && (barf % co))*/
	    printf ("%s\n", fs ? ce : "");
    }


/**** /proc/devices ****/

    if (devicesfp) {
	int maj[2][MAX_DEV];
	char *n[2][MAX_DEV];
	int count[2] =
	{0, 0};
	int which = 0, i, j;

	memset (n, 0, sizeof (n));
	fseek (devicesfp, 0L, SEEK_SET);
	printf ("%s\nCharacter Devices:                      "
		"Block Devices:\n",
		fs ? ce : "");
	while (fgets (line, sizeof (line), devicesfp) && 
			count[CDRV] < MAX_DEV &&
			count[BDRV] < MAX_DEV) {
	    switch (line[0]) {
	    case 'C':
		which = CDRV;
		break;
	    case 'B':
		which = BDRV;
		break;
	    case '\n':
		break;
	    default:
		maj[which][count[which]] =
		    atoi (strtok (line, " "));
		n[which][count[which]] =
		    strdup (strtok (NULL, "\n"));
		count[which]++;
		break;
	    }
	}

	j = (1 + MAX (count[0], count[1])) / 2;
	for (i = 0; i < j; i++) {
	    if (n[CDRV][i]) {
		printf ("%3d %-15s ", maj[CDRV][i], n[CDRV][i]);
		free (n[CDRV][i]);
	    } else
		fputs ("                    ", stdout);
	    if (n[CDRV][i + j]) {
		printf ("%3d %-15s ",
			maj[CDRV][i + j], n[CDRV][i + j]);
		free (n[CDRV][i + j]);
	    } else
		fputs ("                    ", stdout);
	    if (n[BDRV][i]) {
		printf ("%3d %-15s ", maj[BDRV][i], n[BDRV][i]);
		free (n[BDRV][i]);
	    } else
		fputs ("                     ", stdout);
	    if (n[BDRV][i + j]) {
		printf ("%3d %-15s ",
			maj[BDRV][i + j], n[BDRV][i + j]);
		free (n[BDRV][i + j]);
	    }
	    printf ("%s\n", fs ? ce : "");
	    if (i >= count[CDRV] && i >= count[BDRV])
		break;
	}
    }


/**** /proc/filesystems ****/

    if (filesystemsfp) {
	int barf = 0;

	fseek (filesystemsfp, 0L, SEEK_SET);
	printf ("%s\n", fs ? ce : "");
	printf ("File Systems:%s\n", fs ? ce : "");
	while (fgets (line, sizeof (line), filesystemsfp)) {
	    char *fs;
	    char tmp[21];

	    if (co - barf < 20 ) {
		barf = 0;
		putchar('\n');
	    }
	    barf += 20;

	    fs = strchr (line, '\t');
	    fs = strtok (fs + 1, "\n");
	    if (line[0] == 'n') {
		snprintf (tmp, sizeof(tmp), "[%s]", fs);
		printf ("%-20s", tmp);
	    } else
		printf ("%-20s", fs);
	}
	/* if (co && (barf % co)) */
	    printf ("%s\n", fs ? ce : "");
    }
}


int
main (int ac, char **av)
{
    long sl = 5000000L;
    int getoptopt, scanintr;
    struct timeval then, now;
#ifndef DEBUG
    char outbuf[4096];
#endif

    uname(&ut);
    linux_version_code = (atol(strtok(ut.release, ".")) * 0x10000) +
	(atol(strtok(NULL, ".")) * 0x100) +
	atol(strtok(NULL, ""));
    get_HZ();

    while ((getoptopt = getopt (ac, av, "sfn:madirDSF:bChv")) != EOF) {
	switch (getoptopt) {
	case 'n':
	    sl = (long) (atof (optarg) * 1000000.0);
	   /* PLUMMET */
	case 'f':
	    fs = 1;
	    break;
	case 's':
	    show_stat = 1;
	    show_moddev = show_all = 0;
	    break;
	case 'm':
	    show_moddev = 1;
	    break;
	case 'a':
	    show_all = 1;
	    break;
	case 'S':
	    per_sec = 1;
	    break;
	case 'F':
	    if ((freopen (optarg, "r", stdin)) == NULL ||
	        (freopen (optarg, "w", stdout)) == NULL ||
	        (freopen (optarg, "w", stderr)) == NULL) {
		fprintf (stderr, "%s: ", av[0]); /* What happens if the third one failed, anyway? */
		perror ("unable to open new output file");
		exit (errno);
	    }
	    break;
	case 'D':
	    show_new_mem = 1;
	    /* PLUMMET */
	case 'd':
	    show_diff = fs = 1;
	    break;
	case 'i':
	    squeeze_irqs = 0;
	    break;
	case 'r':
	    show_real_mem = 1;
	    break;
	case 'b':
	    io_or_blk = 1;
	    break;
	case 'v':
	    printf ("This is procinfo version " VERSION "\n");
	    exit (0);
	case 'h':
	default:
	    printf ("procinfo version %s\n"
		    "usage: %s [-fsmadiDSbhv] [-nN] [-Ffile]\n"
		    "\n"
		    "\t-s\tdisplay memory, disk, IRQ & DMA info (default)\n"
		    "\t-m\tdisplay module and device info\n"
		    "\t-a\tdisplay all info\n"
		    "\t-f\trun full screen\n"
		    "\n"
		    "\t-i\tshow all IRQ channels, not just those used\n"
		    "\t-nN\tpause N second between updates (implies -f)\n"
		    "\t-d\tshow differences rather than totals (implies -f)\n"
		    "\t-D\tshow current memory/swap usage, differences on rest\n"
		    "\t-S\twith -nN and -d/-D, always show values per second\n"
		    "\t-r\tshow memory usage -/+ buffers/cache\n"
		    "\t-Ffile\tprint output to file -- normally a tty\n"
		    "\t-v\tprint version info\n"
		    "\t-h\tprint this help\n",
		    VERSION, av[0]);
	    exit (getoptopt == 'h' ? 0 : 1);
	}
    }

    if (sl == 0)
	nice (-20);


#ifdef DEBUG_VERSION
    versionfp = myfopen (DEBUG_VERSION "version");
#else
    versionfp = myfopen (PROC_DIR "version");
#endif
    uptimefp = myfopen (PROC_DIR "uptime");
    loadavgfp = myfopen (PROC_DIR "loadavg");
    meminfofp = myfopen (PROC_DIR "meminfo");
    statfp = myfopen (PROC_DIR "stat");
    /* These may be missing, so check for NULL later. */
    diskstatsfp = fopen (PROC_DIR "diskstats", "r");
    vmstatfp = fopen (PROC_DIR "vmstat", "r");
    modulesfp = fopen (PROC_DIR "modules", "r");
    devicesfp = fopen (PROC_DIR "devices", "r");
    filesystemsfp = fopen (PROC_DIR "filesystems", "r");
    interruptsfp = fopen (PROC_DIR "interrupts", "r");
    dmafp = fopen (PROC_DIR "dma", "r");
    cmdlinefp = fopen (PROC_DIR "cmdline", "r");

#ifndef DEBUG
    setvbuf (stdout, outbuf, _IOFBF, sizeof (outbuf));
#endif

    init_terminal_data ();
    if (fs)
	window_init ();
    winsz (0);

    if (fs) {
	cd = my_tgets ("cd");
	ce = my_tgets ("ce");
	cl = my_tgets ("cl");
	cm = my_tgets ("cm");
	ho = my_tgets ("ho");
	se = my_tgets ("se");
	so = my_tgets ("so");
	ve = my_tgets ("ve");
	vi = my_tgets ("vi");
	sg = tgetnum ("sg");	/* Oh god. */

/*
   If you're suffering from terminals with magic cookies, 1) you have
   my sympathy, and 2) you might wish just to forget about standout
   altogether. In that case, define COOKIE_NOSOSE.
 */

#ifdef COOKIE_NOSOSE
	if (sg > 0) {
	    sg = 0;			/* Just forget about se/so */
	    se = so = "";
	}
#endif
	if (sg < 0)
	    sg = 0;
    }

    /* Count number of CPUs */
    nr_cpus = sysconf (_SC_NPROCESSORS_ONLN);

    /* Gets called from winsz(), but in case stdout is redirected: */
    version = make_version (versionfp);

    /* See what the intr line in /proc/stat says. */
    scanintr = 0;
    while (fgets (line, sizeof (line), statfp)) {
	if (scanintr || !strncmp("intr", line, 4)) {
	    /* Counting number of spaces after line[5] gives us */
	    /* number of interrupt entries in new-style intr. */
	    /* If zero: old style intr. */
	    size_t i, len = strlen(line);
	    for(i = ((scanintr) ? 0 : 5); i < len; i++) {
		if (line[i] == '\n')
		    break;
		if (line[i] == ' ')
		    nr_irqs++;
	    }
	    if (line[i] != '\n')
		scanintr = 1;
	    else
		scanintr = 0;
	    continue;
	}
	/* While we're at it, fill in booted. */
	else if (!strncmp ("btime", line, 5)) {
	    time_t btime;

	    btime = (time_t) atol (&line[6]);
	    strftime (booted, sizeof (booted), "%c", localtime (&btime));
	    continue;
	}
    }
    if (nr_irqs) {
	new.intr = my_xcalloc (nr_irqs, sizeof (unsigned long int));
	old.intr = my_xcalloc (nr_irqs, sizeof (unsigned long int));
	memset (&base, 0, sizeof (struct info));
	base.intr = my_xcalloc (nr_irqs, sizeof (unsigned long int));
    }

    if (fs)
	printf ("%s%s", cl, vi);

    gettimeofday (&now, 0);
    while (42) {

	if (fs && redrawn) {
	    redrawn = 0;
	    fputs (cl, stdout);
	}

	if (fs) {
	    printf ("%s%s%*s%s\n", ho, so, -(co - 2 * sg), version, se);
	    printf ("%s", ce);
	} else
	    printf ("%s\n", version);

	if (message) {
	    printf ("[%s]", message);
	    message = NULL;
	}
	putchar ('\n');

	if (show_moddev + show_sys == 0)
	    show_stat = 1;
	if (show_all)
	    show_stat = show_moddev = show_sys = 1;

	if (show_stat)
	    first_page (sl);

	if (show_moddev)
	    second_page ();

	fflush (stdout);

	if (fs) {
	    fd_set readfd;
	    struct timeval tv;
	    int ret;

	    tv.tv_sec = sl / 1000000;
	    tv.tv_usec = sl - (tv.tv_sec * 1000000);

	    FD_ZERO(&readfd);
	    FD_SET(0, &readfd);

	    ret = select (1, &readfd, NULL, NULL, &tv);

	    /* Just ignore any errors. */
	    if (ret > 0) {
		char key;

		ret = read (0, &key, 1);
		switch (key) {

		case 's':
		    show_stat = 1;
		    show_moddev = show_all = 0;
		    redrawn = 1;
		    message = "showing system status";
		    break;
		case 'm':
		    show_moddev = 1;
		    show_stat = show_all = 0;
		    redrawn = 1;
		    message = "showing modules & devices";
		    break;
		case 'a':
		    show_all = show_stat = show_moddev = 1;
		    redrawn = 1;
		    message = "showing all info";
		    break;

		case 't':
		    show_diff = show_new_mem = 0;
		    redrawn = 1;
		    message = "showing totals";
		    break;
		case 'd':
		    show_diff = 1;
		    show_new_mem = 0;
		    redrawn = 1;
		    message = "showing diffs per second/interval";
		    break;
		case 'D':
		    show_diff = show_new_mem = 1;
		    redrawn = 1;
		    message = "showing diffs per second/interval, totals for memory";
		    break;
		case 'S':
		    per_sec = per_sec ? 0 : 1;
		    redrawn = 1;
		    message = per_sec? "showing diffs per second" :
			"showing diffs per interval";
		    break;
		case 'r':
		    show_real_mem = show_real_mem ? 0 : 1;
		    redrawn = 1;
		    message = show_real_mem ?
			"showing memory usage -/+ buffers/cache" :
			    "showing memory usage";
		    break;
		case 'i':
		    squeeze_irqs = squeeze_irqs ? 0 : 1;
		    redrawn = 1;
		    message = squeeze_irqs ? "squeezed IRQ display"
			: "full IRQ display";
		    break;
		case 'b':
		    io_or_blk = io_or_blk ? 0 : 1;
		    /* Fudge things along a little... */
		    new.disk_r[0] = new.disk_w[0] =
			new.disk_r[1] = new.disk_w[1] =
			new.disk_r[2] = new.disk_w[2] =
			new.disk_r[3] = new.disk_w[3] = 0;
			new.disk_r[4] = new.disk_w[4] = 0;
			new.disk_r[5] = new.disk_w[5] = 0;
			new.disk_r[6] = new.disk_w[6] = 0;
		    redrawn = 1;
		    message = io_or_blk ? "showing I/O in blocks" :
			"showing I/O per requests";
		    break;

		case 'n':
		{
		    char line[41];
		    double tmpsl;

		    printf ("delay: ");
		    set_echo (1);
		    fgets (line, 40, stdin);
		    set_echo (0);
		    if (strlen (line) > 1) {
			tmpsl = atof (line);
			sl = (long) (tmpsl * 1000000.0);
			snprintf (line, sizeof(line), "delay set to %f", tmpsl);
			message = line;
		    }
		    redrawn = 1;
		    break;
		}

		case 'C':
		{
		    show_from_baseline = 1;
		    if (nr_irqs)
			COPYINFO(base, new)
			else
			    memcpy (&base, &new, sizeof (struct info));
		    message = "checkpoint set";
		    break;
		}
		case 'R':
		    show_from_baseline = 0;
		    message = "checkpoint released";
		    break;

		case ' ':
		    printf ("[hold]"); fflush(stdout);
		    read (0, &key, 1);
		    printf ("\r      "); fflush(stdout);
		    break;

		case 12:	/* ^L */
		    redrawn = 1;
		    break;

		case 'h':
		case '?':
		    message = "s/m/a   t/d/D   S i b n  <space> ^L ? q";
		    break;
		case 'q':
		    quit (0);
		    break;

		case 0:		/* happens if stdin = /dev/null */
		    break;

		default:
		    putchar (7);
		    message = "unknown command";
		    break;
		}
	    }

	    then = now;
	    gettimeofday (&now, 0);
	    if (per_sec) {
		struct timeval sub;
		timersub(&now, &then, &sub);
		rate = (float)sub.tv_sec+(float)sub.tv_usec/1.0e6;
	    }

	} else {
	    putchar ('\n');
	    exit (0);
	}
	if (nr_irqs) {
	    if (show_from_baseline == 0)
		COPYINFO (old, new)
	}
	else {

	    if (show_from_baseline == 0)
		memcpy (&old, &new, sizeof (struct info));
	}
    }				/* 42 */
}

/*
   Local variables:
   rm-trailing-spaces: t
   End:
*/
